////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1999, Vrije Universiteit Amsterdam and University of Twente.
// Author: Frank Dehne (frank@cs.vu.nl).
// Author: Henk van de Zandschulp (henkz@cs.utwente.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "intersect.h"
#include "shapeview.h"
#include "buildingblock.h"
#include "grafport.h"
#include <stdlib.h>

const int BuildingBlock::BUILDING_BLOCK_WIDTH = 76;
const int BuildingBlock::BUILDING_BLOCK_HEIGHT = 38;
const int BuildingBlock::BUILDING_BLOCK_HANDLES = 8;

BuildingBlock::BuildingBlock(ShapeView *v, Grafport *g, int x, int y, int w, int h): 
		NodeShape(v,g,x,y,w,h) {
	SetHandles(BUILDING_BLOCK_HANDLES);
	Point pt(*GetPosition());
	SetPosition(&pt);
	points = new Point[12];
}

BuildingBlock::BuildingBlock(const BuildingBlock &d): NodeShape(d) {
	int w = (GetWidth() < MIN_WIDTH) ? BUILDING_BLOCK_WIDTH : GetWidth();
        int h = (GetHeight() < MIN_HEIGHT) ? BUILDING_BLOCK_HEIGHT : GetHeight();
	SetWidth(w);
	SetHeight(h);
	points = new Point[12];
	for (int i=0;i<12;i++)
		points[i] = d.points[i];
}

BuildingBlock::BuildingBlock(const NodeShape &d): NodeShape(d) {
	SetResizable(True);
        SetFixedName(False);
        SetFixedIndexLabel(False);
	SetHandles(BUILDING_BLOCK_HANDLES);
	SetWidth(max(GetWidth(), MIN_WIDTH));
        SetHeight(max(GetHeight(), MIN_HEIGHT));
	Point pt(*GetPosition());
	SetPosition(&pt);
	points = new Point[12];
}

BuildingBlock::~BuildingBlock() {
	delete[] points;
}

void BuildingBlock::DrawOutLine(const Point *p){
	int x = GetView()->Snap(p->x);
	int y = GetView()->Snap(p->y);
	CalcPoints(x, y);
	SetOutlineAttributes();
	GetGrafport()->DrawPolygon(points, 12);
}

void BuildingBlock::CalcPoints(int x, int y) {
	int a = GetWidth() / 2;
	int b = GetHeight() / 2;
	int c = GetWidth() / 8;
	int d = GetHeight() / 5;

	points[0].x = x - a;
	points[0].y = y - b;
	points[1].x = x + a;
	points[1].y = y - b;
	points[2].x = x + a;
	points[2].y = y + b;
	points[3].x = x - a;
	points[3].y = y + b;
	points[4].x = x - a;
	points[4].y = y + b - d;
	points[5].x = x - a + c;
	points[5].y = y + b - d;
	points[6].x = x - a + c;
	points[6].y = y + b - 2*d;
	points[7].x = x - a;
	points[7].y = y + b - 2*d;
	points[8].x = x - a;
	points[8].y = y + b - 3*d;
	points[9].x = x - a + c;
	points[9].y = y + b - 3*d;
	points[10].x = x - a + c;
	points[10].y = y + b - 4*d;
	points[11].x = x - a;
	points[11].y = y + b - 4*d;
}

Point BuildingBlock::GiveSnp(Point to) const {
	Point p = Intersect::BoxIntersects(*GetPosition(), to, 
		GetWidth(), GetHeight());
	// check if we hit one of the little boxes.
	if (p.x <= points[0].x && ((p.y <= points[4].y && p.y >= points[7].y))) {
		// connect with this little box.
		Point pos;
		pos.x = points[0].x;
		pos.y = (points[4].y + points[7].y) / 2;
		int wd = 2*(points[5].x - points[4].x);
		int ht = points[4].y - points[7].y;
		return Intersect::BoxIntersects(pos, to, wd, ht);
	}
	else if (p.x <= points[0].x && ((p.y <= points[8].y && p.y >= points[11].y))) {
		// connect with this little box.
		Point pos;
		pos.x = points[0].x;
		pos.y = (points[8].y + points[11].y) / 2;
		int wd = 2*(points[9].x - points[8].x);
		int ht = points[8].y - points[11].y;
		return Intersect::BoxIntersects(pos, to, wd, ht);
	}
	else
		return p;
}

Point BuildingBlock::GiveSnp(Point to, int order, int count, int &code) {
	Point p = Intersect::BoxIntersects(*GetPosition(), to, 
		GetWidth(), GetHeight(), order, count, code);
	// check if we hit one of the little boxes.
        if (p.x <= points[0].x && ((p.y <= points[4].y && p.y >= points[7].y))) {
                // connect with this little box.
                Point pos;
                pos.x = points[0].x;
                pos.y = (points[4].y + points[7].y) / 2;
                int wd = 2*(points[5].x - points[4].x);
                int ht = points[4].y - points[7].y;
                return Intersect::BoxIntersects(pos, to, wd, ht,
			order, count, code);
        }
        else if (p.x <= points[0].x && ((p.y <= points[8].y && p.y >= points[11].y))) {
                // connect with this little box.
                Point pos;
                pos.x = points[0].x;
                pos.y = (points[8].y + points[11].y) / 2;
                int wd = 2*(points[9].x - points[8].x);
                int ht = points[8].y - points[11].y;
                return Intersect::BoxIntersects(pos, to, wd, ht,
			order, count, code);
        }
        else
                return p;
}

void BuildingBlock::DrawShape() {
	Grafport *g = GetGrafport();
	CalcPoints(GetPosition()->x, GetPosition()->y);
	int c = GetWidth() / 8;
	if (GetFillStyle() == FillStyle::FILLED) {
		int d = abs(points[7].y-points[4].y);
		g->SetForegroundColor(GetFillColor());
		g->FillPolygon(points, 12);
		g->FillRectangle(points[7].x-c, points[7].y, 2*c, d);
		g->FillRectangle(points[11].x-c, points[11].y, 2*c, d);
		g->SetForegroundColor(GetColor());
	}
	g->DrawPolygon(points, 12);

 	g->DrawLine(points[4].x, points[4].y, points[4].x-c , points[4].y);
	g->DrawLine(points[4].x-c, points[4].y, points[7].x-c, points[7].y);
	g->DrawLine(points[7].x-c, points[7].y, points[7].x, points[7].y);

 	g->DrawLine(points[8].x, points[8].y, points[8].x-c , points[8].y);
	g->DrawLine(points[8].x-c, points[8].y, points[11].x-c, points[11].y);
	g->DrawLine(points[11].x-c, points[11].y, points[11].x, points[11].y);
}

bool BuildingBlock::ContainsPt(int a, int b) {
	double w = (double)GetWidth()/2;
	double h = (double)GetHeight()/2;
	double x = a - GetPosition()->x;
	double y = b - GetPosition()->y;
	// inside the 4 lines of the diamond OR
	// inside one of the 4 corner handles.
	return( y <= -(h/w)*x + h &&
		y >= -(h/w)*x - h &&
		y <=  (h/w)*x + h &&
		y >=  (h/w)*x - h) ||
		(HitHandles(a, b) != -1);
}
