////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1999, Vrije Universiteit Amsterdam and University of Twente.
// Author: Frank Dehne (frank@cs.vu.nl).
// Author: Henk van de Zandschulp (henkz@cs.utwente.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "intersect.h"
#include "grafport.h"
#include "shapeview.h"
#include "cube.h"

Cube::Cube(ShapeView *v, Grafport *g, int x, int y) :
		Hexagon(v, g, x, y) { 
	CalcPositionName();
}


Cube::Cube(const NodeShape &c): Hexagon(c) {
	CalcPositionName();
}

void Cube::DrawOutLine(const Point* p){
	Hexagon::DrawOutLine(p);
	NodeShape::DrawTextShapes();
	Grafport *g = GetGrafport();
	g->SetLineStyle(LineStyle::DOTTED);
	g->DrawLine(points[0].x+1, points[0].y, points[4].x, points[0].y);
	g->DrawLine(points[4].x, points[0].y, points[2].x-1, points[2].y+1);
	g->DrawLine(points[4].x, points[0].y, points[4].x, points[4].y-1);
}


void Cube::DrawShape() {
	Hexagon::DrawShape();
	Grafport *g = GetGrafport();
	string oldBackground;
        if (GetFillStyle() != FillStyle::UNFILLED) {
        	oldBackground = *g->GetBackgroundColor();
             	g->SetBackgroundColor(GetFillColor());
        }
	g->DrawLine(points[0].x+1, points[0].y, points[4].x, points[0].y);
	g->DrawLine(points[4].x, points[0].y, points[2].x-1, points[2].y+1);
	g->DrawLine(points[4].x, points[0].y, points[4].x, points[4].y-1);
        if (GetFillStyle() != FillStyle::UNFILLED)
                g->SetBackgroundColor(&oldBackground);

}


void Cube::CalcPoints(int x, int y) {
	int a = GetWidth()/2;
        int b = GetHeight()/2;
        int c = GetWidth()/8;
        int d = GetHeight()/8;

        points[0].x = x - a;
        points[0].y = y - b + d;
        points[1].x = x - a + c;
        points[1].y = y - b;
        points[2].x = x + a;
        points[2].y = y - b;
        points[3].x = x + a;
        points[3].y = y + b - d;
        points[4].x = x + a - c;
        points[4].y = y + b;
        points[5].x = x - a;
        points[5].y = y + b;
}


Point Cube::GiveSnp(Point to) const {
	Point pos = *GetPosition();
	Point p = Intersect::BoxIntersects(pos, to, GetWidth(), GetHeight());
	// connect to one of the two diagonal lines
	if ((p.x < points[1].x && p.y < points[0].y)) {
		// define little diamond between points[0] and points[1].
		int wd = 2 * (points[1].x - points[0].x);
		int ht = 2 * (points[0].y - points[1].y);
		pos.x = points[1].x;
		pos.y = points[0].y;
		p = Intersect::DiamondIntersects(pos, to, wd, ht);
		// compensation when 'wrong' side of diamond is crossed.
		if (p.x > points[1].x)
			p = points[1];
		else if (p.y > points[0].y)
			p = points[0];
	}
	else if ((p.x > points[4].x && p.y > points[3].y)) {
		// define little diamond between points[3] and points[4].
		int wd = 2 * (points[3].x - points[4].x);
		int ht = 2 * (points[4].y - points[3].y);
		pos.x = points[4].x;
		pos.y = points[3].y;
		p = Intersect::DiamondIntersects(pos, to, wd, ht);
		// compensation when 'wrong' side of diamond is crossed.
		if (p.x < points[4].x)
			p = points[4];
		else if (p.y < points[3].y)
			p = points[3];
	}
	return p;
}

Point Cube::GiveSnp(Point to, int order, int count, int& code) {
	Point pos = *GetPosition();
	Point p = Intersect::BoxIntersects(pos, to, GetWidth(), GetHeight(),
		order, count, code);
	// connect to one of the two diagonal lines
	if ((p.x < points[1].x && p.y < points[0].y)) {
		// define little diamond between points[0] and points[1].
		int wd = 2 * (points[1].x - points[0].x);
		int ht = 2 * (points[0].y - points[1].y);
		pos.x = points[1].x;
		pos.y = points[0].y;
		p = Intersect::DiamondIntersects(pos, to, wd, ht,
			order, count, code);
		if (p.x > points[1].x)
			p = points[1];
		else if (p.x > points[0].y)
			p = points[0];
	}
	else if ((p.x > points[4].x && p.y > points[3].y)) {
		// define little diamond between points[3] and points[4].
		int wd = 2 * (points[3].x - points[4].x);
		int ht = 2 * (points[4].y - points[3].y);
		pos.x = points[4].x;
		pos.y = points[3].y;
		p = Intersect::DiamondIntersects(pos, to, wd, ht,
			order, count, code);
		// compensation when 'wrong' side of diamond is crossed.
		if (p.x < points[4].x)
			p = points[4];
		else if (p.y < points[3].y)
			p = points[3];
	}
	return p;
}
 
void Cube::SetSize(int w, int h) {
        // if (w > h/3)   int c = GetHeight()/3;  see below
	// w = max(w, h);          we expect at least a square box
	w = max(w, h/3 + 2);
        Hexagon::SetSize(w, h);
	CalcPositionName();
}

void Cube::CalcPositionName() {
        Point pt = *GetPosition();
        pt.x -= GetWidth()/16;
        GetName()->SetPosition(&pt);
}
 
void Cube::SetPosition(const Point *pos, bool b) {
        Hexagon::SetPosition(pos, b);
        CalcPositionName();
}
 
void Cube::SetTextShape() {
        Hexagon::SetTextShape();
        CalcPositionName();
}

