////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1995, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "intersect.h"
#include "shapeview.h"
#include "grafport.h"
#include "disk.h"

const int Disk::DISK_WIDTH = 76;
const int Disk::DISK_HEIGHT = 46;
const int Disk::DISK_HANDLES = 8;
const double Disk::ELLIPSE_FACTOR = 0.3;

Disk::Disk(ShapeView *v, Grafport *g, int x, int y, int w, int h): 
		NodeShape(v, g, x, y, w, h) {
	SetHandles(DISK_HANDLES);
	Point pt(*GetPosition());
	SetPosition(&pt);
}

Disk::Disk(const NodeShape &c): NodeShape(c) {
	SetResizable(True);
        SetFixedName(False);
        SetFixedIndexLabel(False);
	SetHandles(DISK_HANDLES);
	int w = (GetWidth() < MIN_WIDTH) ? DISK_WIDTH : GetWidth();
	int h = (GetHeight() < MIN_HEIGHT) ? DISK_HEIGHT : GetHeight();
	SetWidth(w);
	SetHeight(h);
	Point pt(*GetPosition());
	SetPosition(&pt);
}

void Disk::DrawOutLine(const Point *p){
	int x = GetView()->Snap(p->x);
	int y = GetView()->Snap(p->y);
	x = x - GetWidth() / 2;
	y = y - GetHeight() / 2;
	SetOutlineAttributes();
	int eh = (int)(GetHeight()*ELLIPSE_FACTOR);
	GetGrafport()->DrawDisk(x, y, GetWidth(), GetHeight(), eh);
}

void Disk::DrawShape() {
	int eh = (int)(GetHeight()*ELLIPSE_FACTOR);
	string oldBackground;
	Grafport *g = GetGrafport();
	if (GetFillStyle()==FillStyle::FILLED) {
		GetGrafport()->SetForegroundColor(GetFillColor());
		GetGrafport()->FillDisk(
			GetTopLeft()->x, GetTopLeft()->y, 
			GetWidth(), GetHeight(), eh);
		GetGrafport()->SetForegroundColor(GetColor());
        	oldBackground = *g->GetBackgroundColor();
             	g->SetBackgroundColor(GetFillColor());
	}
	GetGrafport()->DrawDisk(
		GetTopLeft()->x, GetTopLeft()->y, 
		GetWidth(), GetHeight(), eh);
	if (GetFillStyle() == FillStyle::FILLED)
                g->SetBackgroundColor(&oldBackground);
}

Point Disk::GiveSnp(Point to) const {
	int eh = (int)(GetHeight()*ELLIPSE_FACTOR);
	Point pos1 = *GetPosition();
	Point pos2 = Point(pos1.x, pos1.y - GetHeight()/2 + eh/2);
	Point pos3 = Point(pos1.x, pos1.y + GetHeight()/2 - eh/2);
        Point p1 = Intersect::BoxIntersects(pos1, to, GetWidth(), GetHeight());
        Point p2 = Intersect::EllipseIntersects(pos2, to, GetWidth(), eh);
        Point p3 = Intersect::EllipseIntersects(pos3, to, GetWidth(), eh);
	if (p2.y < pos2.y && p1.y < pos2.y)
		return p2;
	else if (p3.y > pos3.y && p1.y > pos3.y)
		return p3;
	else
		return p1;
}

Point Disk::GiveOrthoSnp(Point to) const {
	int eh = (int)(GetHeight()*ELLIPSE_FACTOR);
	Point pos1 = *GetPosition();
	Point pos2 = Point(pos1.x, pos1.y - GetHeight()/2 + eh/2);
	Point pos3 = Point(pos1.x, pos1.y + GetHeight()/2 - eh/2);
        Point p1 = GiveBoxOrthoSnp(to);
        Point p2 = Intersect::EllipseIntersects(pos2, to, GetWidth(), eh);
        Point p3 = Intersect::EllipseIntersects(pos3, to, GetWidth(), eh);
	if (p2.y < pos2.y && p1.y < pos2.y)
		return p2;
	else if (p3.y > pos3.y && p1.y > pos3.y)
		return p3;
	else
		return p1;
}

Point Disk::GiveSnp(Point to, int o, int n, int &c) {
	int eh = (int)(GetHeight()*ELLIPSE_FACTOR);
	Point pos1 = *GetPosition();
	Point pos2 = Point(pos1.x, pos1.y - GetHeight()/2 + eh/2);
	Point pos3 = Point(pos1.x, pos1.y + GetHeight()/2 - eh/2);
        Point p1 = Intersect::BoxIntersects(pos1, to, GetWidth(), GetHeight(), o, n, c);
        Point p2 = Intersect::EllipseIntersects(pos2, to, GetWidth(), eh, o, n, c);
        Point p3 = Intersect::EllipseIntersects(pos3, to, GetWidth(), eh, o, n, c);
	if (p2.y < pos2.y && p1.y < pos2.y)
		return p2;
	else if (p3.y > pos3.y && p1.y > pos3.y)
		return p3;
	else
		return p1;
}

Point Disk::GiveBoxOrthoSnp(Point to) const {
        Point pt;
        Point pos = *GetPosition();
        Point tl = *GetTopLeft();
        if (tl.x < to.x && to.x < tl.x + GetWidth()) {
                pt.x = to.x;
                if (to.y < pos.y)
                        pt.y = tl.y;
                else
                        pt.y = tl.y + GetHeight();
                return pt;
        }
        else if (tl.y < to.y && to.y < tl.y + GetHeight()) {
                pt.y = to.y;
                if (to.x < pos.x)
                        pt.x = tl.x;
                else
                        pt.x = tl.x + GetWidth();
                return pt;
        }
        return Intersect::BoxIntersects(pos, to, GetWidth(), GetHeight());
}

void Disk::SetSize(int w, int h) {
        NodeShape::SetSize(w, h);
        CalcPositionName();
}
 
void Disk::SetPosition(const Point *pos, bool b) {
        NodeShape::SetPosition(pos, b);
        CalcPositionName();
}
 
void Disk::SetTextShape() {
        NodeShape::SetTextShape();
        CalcPositionName();
}

void Disk::CalcPositionName() {
        Point pt = *GetPosition();
        pt.y += GetHeight()/6;
        GetName()->SetPosition(&pt);
}
