////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1995, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "intersect.h"
#include "grafport.h"
#include "textshape.h"
#include "shapeview.h"
#include "diagramviewer.h"
#include <math.h>
#include "ellipse.h"

const int Ellipse::ELLIPSE_WIDTH = 76;
const int Ellipse::ELLIPSE_HEIGHT = 38;
const int Ellipse::ELLIPSE_HANDLES = 8;

Ellipse::Ellipse(ShapeView *v, Grafport *g, int x, int y, int w, int h): 
		NodeShape(v, g, x, y, w, h){
	SetHandles(ELLIPSE_HANDLES);
	Point pt(*GetPosition());
	SetPosition(&pt);
}

Ellipse::Ellipse(const NodeShape &c): NodeShape(c) {
	SetResizable(True);
        SetFixedName(False);
        SetFixedIndexLabel(False);
	SetHandles(ELLIPSE_HANDLES);
	int w = (GetWidth() < MIN_WIDTH) ? ELLIPSE_WIDTH : GetWidth(); 
	int h = (GetHeight() < MIN_HEIGHT) ? ELLIPSE_HEIGHT : GetHeight();
        SetWidth(w);
        SetHeight(h);
	Point pt(*GetPosition());
	SetPosition(&pt);
}

void Ellipse::DrawOutLine(const Point *p){
	int x = GetView()->Snap(p->x);
	int y = GetView()->Snap(p->y);
	x = x - GetWidth() / 2;
	y = y - GetHeight() / 2;
	SetOutlineAttributes();
	GetGrafport()->DrawEllipse(x, y, GetWidth(), GetHeight());
}

void Ellipse::DrawDuplicate() {
	int x = GetPosition()->x - 9*GetWidth()/20;
	int y = GetPosition()->y - 2*GetHeight()/5;
	GetGrafport()->DrawStringCentered(x, y, "*");
}

void Ellipse::AdjustSize() {
	// update size of ellipse to fit the textshape.
	// first test if text fits ellipse.
	int twidth, theight;
	GetName()->TextBox(twidth, theight);
	double tw = twidth/2;
	double th = theight/2;
	double wd = GetWidth()/2;
	double ht = GetHeight()/2;
	double fit = (tw*tw)/(wd*wd) + (th*th)/(ht*ht);
	if (fit <= 1)
		return;
	// fit text in ellipse keeping w_new/h_new = wd/ht.
	// w_new derived from ellipse equation:
	// tw^2 / w^2 + th^2 / h^2 = 1.
	double w_new = 2 * sqrt ( (tw*tw) + (th*th*wd*wd)/(ht*ht)); 
	double h_new = w_new * ht/wd;
	UpdateSize((int)(w_new+0.5), (int)(h_new+0.5));
	GetViewer()->ShapePositionUpdate(this);
}

Point Ellipse::GiveSnp(Point to) const {
	return Intersect::EllipseIntersects(*GetPosition(), to,
		GetWidth(), GetHeight());
}

Point Ellipse::GiveSnp(Point to, int order, int count, int &code) {
	return Intersect::EllipseIntersects(*GetPosition(), to,
		GetWidth(), GetHeight(), order, count, code);
}

void Ellipse::DrawShape() {
	if (GetFillStyle() == FillStyle::FILLED) {
		GetGrafport()->SetForegroundColor(GetFillColor());
		GetGrafport()->FillEllipse(
			GetTopLeft()->x, GetTopLeft()->y, 
			GetWidth(), GetHeight());
		GetGrafport()->SetForegroundColor(GetColor());
	}
	GetGrafport()->DrawEllipse(
		GetTopLeft()->x, GetTopLeft()->y, 
		GetWidth(), GetHeight());
}
