////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1997, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "intersect.h"
#include "grafport.h"
#include "shapeview.h"
#include "hexagon.h"

const int Hexagon::HEXAGON_WIDTH = 76;
const int Hexagon::HEXAGON_HEIGHT = 38;
const int Hexagon::HEXAGON_HANDLES = 8;

Hexagon::Hexagon(ShapeView *v, Grafport *g, int x, int y, int w, int h): 
		NodeShape(v,g,x,y,w,h) {
	SetHandles(HEXAGON_HANDLES);
	Point pt(*GetPosition());
	SetPosition(&pt);
	points = new Point[6];
}

Hexagon::~Hexagon() {
	delete[] points;
}

Hexagon::Hexagon(const Hexagon &h): NodeShape(h) {
	points = new Point[6];
	for (int i=0; i<6; i++)
		points[i] = h.points[i];
} 

Hexagon::Hexagon(const NodeShape &h): NodeShape(h) {
	SetResizable(True);
        SetFixedName(False);
        SetFixedIndexLabel(False);
	SetHandles(HEXAGON_HANDLES);
	int wd = (GetWidth() < MIN_WIDTH) ? HEXAGON_WIDTH : GetWidth();
        int ht = (GetHeight() < MIN_HEIGHT) ? HEXAGON_HEIGHT : GetHeight();
        SetWidth(wd);
        SetHeight(ht);
	Point pt(*GetPosition());
	SetPosition(&pt);
	points = new Point[6];
}

void Hexagon::DrawOutLine(const Point *p){
	int x = GetView()->Snap(p->x);
	int y = GetView()->Snap(p->y);
	CalcPoints(x, y);
	SetOutlineAttributes();
	GetGrafport()->DrawPolygon(points, 6);
}

void Hexagon::CalcPoints(int x, int y) {
	int a = GetWidth()/4;
	int b = GetHeight()/2;
	points[0].x = x - a;
	points[0].y = y - b;
	points[1].x = x + a;
	points[1].y = y - b;
	points[2].x = x + 2*a;
	points[2].y = y;
	points[3].x = x + a;
	points[3].y = y + b;
	points[4].x = x - a;
	points[4].y = y + b;
	points[5].x = x - 2*a;
	points[5].y = y;
}

Point Hexagon::GiveSnp(Point to) const {
	Point pos = *GetPosition();
	Point p = Intersect::BoxIntersects(pos, to, GetWidth(), GetHeight());
	if (p.x < pos.x - GetWidth()/4 || p.x > pos.x + GetWidth()/4)
		// regard hexagon as twice as large diamond.
		return Intersect::DiamondIntersects(pos, to, GetWidth(), 
				2*GetHeight());
	else 
		return p;
}

Point Hexagon::GiveSnp(Point to, int order, int count, int &code) {
	Point pos = *GetPosition();
	Point p = Intersect::BoxIntersects(pos, to, GetWidth(), GetHeight(), 
			order, count, code);
	if (p.x < pos.x - GetWidth()/4 || p.x > pos.x + (GetWidth()+2)/4) {
		// regard hexagon as a diamond that is moved and twice as small.
		Point q = pos; 
		if (p.x < q.x)
			q.x -= GetWidth()/4; 
		else
			q.x += GetWidth()/4; 
		return Intersect::DiamondIntersects(q, to, 
			GetWidth()/2, GetHeight(), order, count, code);
	}
	else
		return p;
}

void Hexagon::DrawShape(){
	Point pos = *GetPosition();
	CalcPoints(pos.x, pos.y);
	if (GetFillStyle() == FillStyle::FILLED) {
		GetGrafport()->SetForegroundColor(GetFillColor());
		GetGrafport()->FillPolygon(points, 6);
		GetGrafport()->SetForegroundColor(GetColor());
	}
	GetGrafport()->DrawPolygon(points, 6);
}

bool Hexagon::ContainsPt(int a, int b) {
	return Shape::ContainsPt(a, b);
}
