////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1996, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "grafport.h"
#include "intersect.h"
#include "textshape.h"
#include "subject.h"
#include "lineend.h"
#include "miniarrowellipse.h"

const int MiniArrowEllipse::ARROWLEN = 20;
const int MiniArrowEllipse::CIRCLE_WIDTH = 12;

MiniArrowEllipse::MiniArrowEllipse(ShapeView *v, Grafport *g, int x, int y,
		int w, int h): Ellipse(v, g, x, y, w, h) {
	Init();
}
 
MiniArrowEllipse::MiniArrowEllipse(const MiniArrowEllipse &c): Ellipse(c) {
	textHeight = c.textHeight;
	CalcPositionName();
}

MiniArrowEllipse::MiniArrowEllipse(const NodeShape &c): Ellipse(c) {
	Init();
}
 
void MiniArrowEllipse::Init() {
	textHeight = GetName()->GetFontHeight();
	int extra = ARROWLEN + textHeight;
	Shape::SetSize(max(GetWidth(), GetName()->GetWidth()), 
			   GetHeight()+extra);
	Point pt = *GetPosition();
	pt.y -= extra/2;
	SetPosition(&pt);
	CalcPositionName();
	SetHandles(MIN_HANDLES);
	SetFixedName(True);
	SetFixedIndexLabel(True);
	end.SetType(LineEnd::FILLED_ARROW);
}

void MiniArrowEllipse::SetPosition(const Point *p, bool snap) {
	Ellipse::SetPosition(p, snap);
	CalcPositionName();
}

void MiniArrowEllipse::SetFont(XFont *ft) {
	Ellipse::SetFont(ft);
	CalcPositionName();
}

void MiniArrowEllipse::SetAlignment(TextAlign::Type al) {
	Ellipse::SetAlignment(al);
	CalcPositionName();
}

void MiniArrowEllipse::SetName(const string *s) {
	GetName()->SetString(s);
	Shape::SetSize(max(CIRCLE_WIDTH, GetName()->GetWidth()),
		GetHeight()); 
	CalcPositionName();
}

void MiniArrowEllipse::CalcPositionName() {
	Point pt;
	pt.x = GetPosition()->x;
	pt.y = GetTopLeft()->y + textHeight/2;
	GetName()->SetPosition(&pt);
}
 
TextShape *MiniArrowEllipse::ChooseTextShape(int, int) {
	return GetName();
}

Point MiniArrowEllipse::GiveSnp(Point to) const {
	Point pt = *GetPosition();
	pt.y = GetTopLeft()->y+ARROWLEN+textHeight+CIRCLE_WIDTH/2;
	return Intersect::EllipseIntersects(pt, to, 
			CIRCLE_WIDTH,CIRCLE_WIDTH);
}
 
Point MiniArrowEllipse::GiveSnp(Point to, int order, int count, int &code) {
	Point pt = *GetPosition();
	pt.y = GetTopLeft()->y+ARROWLEN+textHeight+CIRCLE_WIDTH/2;
	return Intersect::EllipseIntersects(pt, to, CIRCLE_WIDTH, 
				CIRCLE_WIDTH, order, count, code);
}

void MiniArrowEllipse::DrawShape() {
	int x = GetPosition()->x;
	int y = GetTopLeft()->y;
	Point from = Point(x, y+textHeight);
	Point to = Point(x, y+textHeight+ARROWLEN-2);
	GetGrafport()->DrawEllipse(x-CIRCLE_WIDTH/2, 
		y+GetHeight()-CIRCLE_WIDTH, CIRCLE_WIDTH, CIRCLE_WIDTH);
	Point newTo;
	end.Draw(GetGrafport(), &from, &to, &newTo);
	GetGrafport()->DrawLine(&from, &newTo);
}

void MiniArrowEllipse::SetTextShape() {
	Ellipse::SetTextShape();
	if (check(GetSubject()))
		GetName()->SetString(GetSubject()->GetName());
	Shape::SetSize(max(CIRCLE_WIDTH, GetName()->GetWidth()),
		GetHeight()); 
	GetName()->SetParent(this);
	CalcPositionName();
}

bool MiniArrowEllipse::HasString(const string *s, bool sens, bool sub) const {
	return GetName()->HasString(s, sens, sub);
}

bool MiniArrowEllipse::HasString(const string *s,
			bool sens, bool sub, List<TextShape *> *list) {
	return GetName()->HasString(s, sens, sub, list);
}
