////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1997, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "grafport.h"
#include "nodeshape.h"
#include "textshape.h"
#include "diagramviewer.h"
#include "node.h"
#include "inputfile.h"
#include "outputfile.h"

NodeShape::NodeShape(ShapeView *v, Grafport *g, int x, int y, int w, int h): 
		GShape(v, g, x, y) {
	SetWidth(w);
	SetHeight(h);
	SetFillStyle(GetViewer()->GetDefaultFillStyle());
	string s = *GetViewer()->GetDefaultFillColor();
	SetFillColor(&s);
	s = *GetViewer()->GetDefaultLineColor();
	SetColor(&s);
	SetLineStyle(GetViewer()->GetDefaultNodeLineStyle());
	indexLabel = new TextShape(v, g, this);
	indexLabel->SetOneLine(True);
	indexLabel->SetDescription("Index");
	fixedIndexLabel = True;
	resizable = True;
	CalcPositionIndexLabel();
}

NodeShape::NodeShape(const NodeShape &c): GShape(c) {
	indexLabel = new TextShape(*(c.indexLabel));
	indexLabel->SetParent(this);
	fixedIndexLabel = c.fixedIndexLabel;
	resizable = c.resizable;
	fillColor = c.fillColor;
	fillStyle = c.fillStyle;
	CalcPositionIndexLabel();
}
 
NodeShape::~NodeShape() {
	delete indexLabel;
}

void NodeShape::DrawTextShapes() {
	// set background color.
	string oldBackground;
	if (GetFillStyle() != FillStyle::UNFILLED) {
		oldBackground = *GetGrafport()->GetBackgroundColor();
		GetGrafport()->SetBackgroundColor(GetFillColor());
	}
	GShape::DrawTextShapes();
	indexLabel->Draw();
	if (IsDuplicate())
                DrawDuplicate();
	// restore background color.
	if (GetFillStyle() != FillStyle::UNFILLED)
		GetGrafport()->SetBackgroundColor(&oldBackground);
}

void NodeShape::UpdateFillColor(const string *c) {
	if (GetFillStyle() == FillStyle::UNFILLED)
		SetFillColor(c);
	else {
		Undraw();
		SetFillColor(c);
		Draw();
	}
}
 
void NodeShape::UpdateFillStyle(FillStyle::Type f) {
	if (GetFillStyle() != f) {
		Undraw();
		SetFillStyle(f);
		Draw();
	}
}

void NodeShape::SetPosition(const Point *pos, bool snap) {
	Shape::SetPosition(pos, snap);
	GetName()->SetPosition(GetPosition());
	CalcPositionIndexLabel();
}

void NodeShape::SetSize(int w, int h) {
	if (!resizable)
		return;
	GShape::SetSize(w, h);
	CalcPositionIndexLabel();
}

void NodeShape::SetFont(XFont *ft) {
	GShape::SetFont(ft);
	indexLabel->SetFont(ft);
}

void NodeShape::SetTextColor(const string *c) {
	GShape::SetTextColor(c);
	indexLabel->SetColor(c);
}

void NodeShape::SetGrafport(Grafport *g) {
	GShape::SetGrafport(g);
	indexLabel->SetGrafport(g);
}
 
void NodeShape::SetView(ShapeView *v) {
	GShape::SetView(v);
	indexLabel->SetView(v);
}

void NodeShape::CalcPositionIndexLabel() {
	Point pt;
	pt.x = GetPosition()->x;
	pt.y = GetPosition()->y - GetHeight()/3;
	indexLabel->SetPosition(&pt);
}

void NodeShape::UpdateIndexLabel(const string *s) {
	if (IsVisible())
		Undraw();
	indexLabel->SetString(s);
	if (IsVisible())
		Draw();
}
 
void NodeShape::DrawDuplicate() {
	int x = GetTopLeft()->x + 6;
	int y = GetTopLeft()->y + 7;
	GetGrafport()->SetFont(GetName()->GetFont());
	GetGrafport()->DrawStringCentered(x, y, "*");
}

void NodeShape::DrawMultiplicity() {
	int x = GetTopLeft()->x + GetWidth() - 6;
	int y = GetTopLeft()->y + 7;
	GetGrafport()->SetFont(GetName()->GetFont());
	GetGrafport()->DrawStringCentered(x, y, "2");
}

TextShape *NodeShape::ChooseTextShape(int x, int y) {
	if (GetViewer()->IsShowIndexes() && !fixedIndexLabel) {
		if (y <= GetPosition()->y - GetHeight()/3 + 5)
			return GetIndexLabel();
	}
	return GShape::ChooseTextShape(x, y);
}

bool NodeShape::InTextArea(int x, int y) {
	if (IsFixedName())
		return False;
	else if (GetName()->InTextArea(x, y))
		// (x,y) in name shape.
		return True;
	else {
		// or (x.y) in node shape itself ?
		Rectangle rect(
			GetTopLeft()->x + HANDLE_WIDTH/2, 
			GetTopLeft()->y + HANDLE_WIDTH/2,
			GetWidth()  - HANDLE_WIDTH,
			GetHeight() - HANDLE_WIDTH);
		return rect.Inside(x, y);
	}
}

void NodeShape::AdjustSize() {
	// update size of gshape (does not work when it is a line) to
	// fit the name textshape (only box-like shapes work well of course,
	// so non-box classes should have a different implementation).
	if (!IsLine()) {
		int twidth, theight;
		GetName()->TextBox(twidth, theight);
		if (twidth > GetWidth() || theight > GetHeight()) {
			UpdateSize(max(twidth+4, GetWidth()),
				   max(theight+4, GetHeight()));
			GetViewer()->ShapePositionUpdate(this);
		}
	}
}

int NodeShape::HitHandles(int x, int y) {
	if (!resizable)
		return -1;
	else
		return GShape::HitHandles(x, y);
}

bool NodeShape::SetAssocSubject(AssocList *al) {
	if (!Shape::SetAssocSubject(al))
		return False;
	if (check(GetSubject() && !GetSubject()->IsEdge()))
		return True;
	SetSubject(0);
	return False;
}

bool NodeShape::ReadMembers(InputFile *ifile, double format) {
	if (!Shape::ReadMembers(ifile, format))
		return False;
	if (!ReadGraphicsPart(ifile, format))
		return False;
	if (format >= 1.21) {
		string val1;
		if (!ifile->ReadAttribute("FillStyle", &val1))
			return False;
		fillStyle = FillStyle::String2Type(&val1);
		if (!ifile->ReadStringAttribute("FillColor", &fillColor))
			return False;
	}
	if (!ReadTextPart(ifile, format))
		return False;
	GetName()->SetPosition(GetPosition());
	return True;
}

void NodeShape::WriteMembers(OutputFile *ofile) {
	Shape::WriteMembers(ofile);
	WriteGraphicsPart(ofile);
	string fs;
	FillStyle::Type2String(fillStyle, &fs);
	(*ofile) << "\t{ FillStyle " << fs  << " }\n";
	(*ofile) << "\t{ FillColor " <<  '"' << fillColor << '"' << " }\n";
	WriteTextPart(ofile);
}

void NodeShape::SetTextShape() {
	GShape::SetTextShape();
	// set position of top label.
	if (GetViewer()->IsShowIndexes())
		if (check(GetSubject() && !GetSubject()->IsEdge()))
			GetIndexLabel()->
				SetString(((Node *)GetSubject())->GetIndex());
	else {
		string empty("");
		GetIndexLabel()->SetString(&empty);
	}
	GetIndexLabel()->SetParent(this);
	CalcPositionIndexLabel();
}

bool NodeShape::HasTextShape(TextShape *t) const {
	if (t == GetName())
		return True;
	else if (t == indexLabel)
		return True;
	else
		return False;
}
 
bool NodeShape::HasString(const string *s, bool sens, bool sub) const {
	if (GShape::HasString(s, sens, sub))
		return True;
	else if (*s != "" && indexLabel->HasString(s, sens, sub))
		return True;
	else
		return False;
}
 
bool NodeShape::HasString(const string *s, bool sens, bool sub,
		List<TextShape *> *list) {
	bool b= GShape::HasString(s, sens, sub, list);
	if (*s != "")
		b += indexLabel->HasString(s, sens, sub, list);
	return b;
}

