////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1999, Vrije Universiteit Amsterdam and University of Twente.
// Author: Frank Dehne (frank@cs.vu.nl).
// Author: Henk van de Zandschulp (henkz@cs.utwente.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "intersect.h"
#include "grafport.h"
#include "textshape.h"
#include "shapeview.h"
#include "diagramviewer.h"
#include <math.h>
#include "stickman.h"

const int StickMan::STICKMAN_WIDTH = 38;
const int StickMan::STICKMAN_HEIGHT = 64;
const int StickMan::STICKMAN_HANDLES = 8;

StickMan::StickMan(ShapeView *v, Grafport *g, int x, int y, int w, int h): 
		NodeShape(v, g, x, y, w, h){
        SetWidth(STICKMAN_WIDTH);
        SetHeight(STICKMAN_HEIGHT);
	SetResizable(True);
	SetHandles(STICKMAN_HANDLES);
	Point pt(*GetPosition());
	SetPosition(&pt);
	CalcPositionLabels();
}

StickMan::StickMan(const NodeShape &c): NodeShape(c) {
	SetResizable(True);
        SetFixedName(False);
        SetFixedIndexLabel(False);
	SetHandles(STICKMAN_HANDLES);
	SetSize(STICKMAN_WIDTH, RequiredHeight());
}

void StickMan::SetPosition(const Point *pos, bool snap) {
	NodeShape::SetPosition(pos, snap);
	CalcPositionLabels();
}

void StickMan::SetSize(int w, int h) {
        NodeShape::SetSize(w, h);
	CalcPositionLabels();
}

void StickMan::CalcPositionLabels() {
	int tht = TextHeight();
	// set name position at the bottom.
	Point pt(*GetPosition());
        pt.y += GetHeight()/2 - tht/2 - 4;
        GetName()->SetPosition(&pt);
	// set index label position.
	pt.x = GetPosition()->x + GetWidth()*5/12;
        pt.y = GetPosition()->y - GetHeight()/3;
        GetIndexLabel()->SetPosition(&pt);
}
 
void StickMan::AdjustSize() {
        int tht = RequiredHeight();
        if (tht != GetHeight()) {
                int ht = max(tht, STICKMAN_HEIGHT);
                UpdateSize(STICKMAN_WIDTH, ht);
                GetViewer()->ShapePositionUpdate(this);
        }
	else
		CalcPositionLabels();
}

int StickMan::TextHeight() {
        int textHeight = GetName()->GetFontHeight();
        int labelLines = GetName()->GetString()->lines();
        return labelLines>0 ? labelLines*textHeight : 0;
}

int StickMan::RequiredHeight() {
        return (STICKMAN_HEIGHT + TextHeight());
}

void StickMan::DrawOutLine(const Point *p){
	int x = GetView()->Snap(p->x);
	int y = GetView()->Snap(p->y);
	x = x - GetWidth() / 2;
	y = y - GetHeight() / 2;
	SetOutlineAttributes();
	DrawStickMan(x, y, GetWidth(), GetHeight()-TextHeight());
	GetGrafport()->DrawRectangle(x, y, GetWidth(), GetHeight());
}

void StickMan::SetTextShape() {
	NodeShape::SetTextShape();
	CalcPositionLabels();
}

void StickMan::DrawDuplicate() {
	int x = GetPosition()->x - 9*GetWidth()/20;
	int y = GetPosition()->y - 2*GetHeight()/5;
	GetGrafport()->DrawStringCentered(x, y, "*");
}

Point StickMan::GiveSnp(Point to) const {
	return Intersect::BoxIntersects(*GetPosition(), to,
		GetWidth(), GetHeight());
}

Point StickMan::GiveSnp(Point to, int order, int count, int &code) {
	return Intersect::BoxIntersects(*GetPosition(), to,
		GetWidth(), GetHeight(), order, count, code);
}

void StickMan::DrawShape() {
	int x = GetTopLeft()->x; 
	int y = GetTopLeft()->y;
	int wd = GetWidth();
	int ht = GetHeight() - TextHeight();
	if (GetFillStyle() == FillStyle::FILLED) {
		GetGrafport()->SetForegroundColor(GetFillColor());
		FillStickMan(x, y, wd, ht);
		GetGrafport()->SetForegroundColor(GetColor());
	}
	DrawStickMan(x, y, wd, ht);
}

void StickMan::DrawStickMan(int x, int y, int wd, int ht) {
	int headSize = min(wd/3, ht/3);
	int head_x = x + wd/2 - headSize/2;
	int head_y = y + 6;
	GetGrafport()->DrawEllipse(head_x, head_y, headSize, headSize);
	int neck_y = head_y + headSize;
	GetGrafport()->DrawLine(x+wd/2, neck_y, x+wd/2, neck_y+ht/3);
	GetGrafport()->DrawLine(x+wd/4, neck_y+ht/12, x+wd-wd/4, neck_y+ht/12);
	int leg_y = neck_y+ht/3;
	GetGrafport()->DrawLine(x+wd/2-1, leg_y, x+wd/4, max(leg_y, y+ht-10));
	GetGrafport()->DrawLine(x+wd/2+1, leg_y, x+wd-wd/4, max(leg_y, y+ht-10));
}

void StickMan::FillStickMan(int x, int y, int wd, int ht) {
	int headSize = min(wd/3, ht/3);
	int head_x = x + wd/2 - headSize/2;
	int head_y = y + 6;
	GetGrafport()->FillEllipse(head_x, head_y, headSize, headSize);
}

 
void StickMan::DrawTextShapes() {
	// do not set background color.
        GShape::DrawTextShapes();
        GetIndexLabel()->Draw();
        if (IsDuplicate())
                DrawDuplicate();
}

