////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1995, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "grafport.h"
#include "intersect.h"
#include "triangle.h"
#include "textshape.h"
#include "shapeview.h"

const int Triangle::TRIANGLE_WIDTH = 60;
const int Triangle::TRIANGLE_HEIGHT = 50;
const int Triangle::TRIANGLE_HANDLES = 6;

Triangle::Triangle(ShapeView *v, Grafport *g, int x, int y, int w, int h): 
		NodeShape(v, g, x, y, w, h) {
	SetHandles(TRIANGLE_HANDLES);
	Point pt(*GetPosition());
	SetPosition(&pt);
}

Triangle::Triangle(const Triangle &t): NodeShape(t) {
	points[0] = t.points[0];
	points[1] = t.points[1];
	points[2] = t.points[2];
	CalcPositionName();
}
 
Triangle::Triangle(const NodeShape &t): NodeShape(t) {
	SetResizable(True);
        SetFixedName(False);
        SetFixedIndexLabel(False);
	SetHandles(TRIANGLE_HANDLES);
	int w = (GetWidth() < MIN_WIDTH) ? TRIANGLE_WIDTH : GetWidth();
        int h = (GetHeight() < MIN_HEIGHT) ? TRIANGLE_HEIGHT : GetHeight();
        SetWidth(w);
        SetHeight(h);
	CalcPoints(GetPosition()->x, GetPosition()->y);
	CalcPositionName();
	Point pt(*GetPosition());
	SetPosition(&pt);
}

void Triangle::DrawDuplicate() {
	int x = GetPosition()->x - GetWidth()/4;
	int y = GetPosition()->y - 2*GetHeight()/5;
	GetGrafport()->DrawStringCentered(x, y, "*");
}
 
void Triangle::DrawOutLine(const Point *p) {
	int x = GetView()->Snap(p->x);
	int y = GetView()->Snap(p->y);
	CalcPoints(x, y);
	SetOutlineAttributes();
	GetGrafport()->DrawPolygon(points, 3);
}

Point Triangle::GiveSnp(Point to) const { 
	return Intersect::TriangleIntersects(*GetPosition(), to, 
		GetWidth(), GetHeight());
}

Point Triangle::GiveSnp(Point to, int order, int count, int &code) {
	return Intersect::TriangleIntersects(*GetPosition(), to,
		GetWidth(), GetHeight(), order, count, code);
}

void Triangle::CalcPoints(int x, int y) {
	int w = GetWidth()/2;
	int h = GetHeight()/2;
	points[0].x = x; 
	points[0].y = y - h; 
	points[1].x = x - w; 
	points[1].y = y + h; 
	points[2].x = x + w; 
	points[2].y = y + h; 
}
 
void Triangle::SetSize(int w, int h) {
	NodeShape::SetSize(w, h);
	CalcPositionName();
}

void Triangle::SetPosition(const Point *pos, bool b) {
	NodeShape::SetPosition(pos, b);
	CalcPositionName();
}

void Triangle::CalcPositionName() {
	Point pt = *GetPosition();
	pt.y += GetHeight()/6;
	GetName()->SetPosition(&pt);
}

void Triangle::SetTextShape() {
	NodeShape::SetTextShape();
	CalcPositionName();
}

void Triangle::DrawShape() {
	CalcPoints(GetPosition()->x, GetPosition()->y);
	if (GetFillStyle() == FillStyle::FILLED) {
		GetGrafport()->SetForegroundColor(GetFillColor());
		GetGrafport()->FillPolygon(points, 3);
		GetGrafport()->SetForegroundColor(GetColor());
	}
	GetGrafport()->DrawPolygon(points, 3);
}

bool Triangle::ContainsPt(int a, int b) {
	double w = (double)GetWidth()/2;
	double h = (double)GetHeight()/2;
	double x = a - GetPosition()->x;
	double y = b - GetPosition()->y;
	// inside the 3 lines of the triangle OR
	// inside one of the 4 corner handles.
	return( y <= h &&
		y >= ((-2*h/w)*x - h) &&
		y >= (( 2*h/w)*x - h)) ||
		(HitHandles(a, b) != -1);
}
