////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1999, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "figgrafport.h"
#include "application.h"
#include "system.h"
#include "intersect.h"
#include <ctype.h>
#include <math.h>
#include <X11/Xlib.h>

// Factor for X coordinates to Fig coordinates.
// X display : 1 dot = 1/80 inch
// X Fig     : 1 dot = 1/1200 inch
// const double FigGrafport::F = 1200/80;
// const double FigGrafport::F = 14.4578;
const double FigGrafport::F = 15;

FigGrafport::FigGrafport(const char *fileName): Grafport() {
	if (equal(fileName, ""))
		fd = stdout;
	else
		fd = fopen(fileName, "w");
	succes = (fd != 0);
	penColor = -1;
	fillColor = -1;
	psFont = -1;
	texFont = 0;
	texFonts = False;
	InitColors();
}

FigGrafport::~FigGrafport() {
	if (succes)
		fclose(fd);
}

void FigGrafport::InitColors() {
	// add standard Xfig colors.
	// "pixel"-value is misused as xfig color key.
	AddColor("black", 0);
	AddColor("blue", 1);
	AddColor("green", 2);
	AddColor("cyan", 3);
	AddColor("red", 4);
	AddColor("magenta", 5);
	AddColor("yellow", 6);
	AddColor("white", 7);
	// here could come standard colors 8-30 (is not necessary).
	AddColor("gold", 31);

	int n = 32; // user defined colors are from 32 till 543.
	List<Color *> *appColors = theApplication->GetColors();
        for (appColors->first(); !appColors->done(); appColors->next()) {
		// define other colors as user defined colors.
		if (!LookupColor(&appColors->cur()->name)) {
			Color *color = new Color(*appColors->cur());
                	color->pixel = n; 
                	color->allocated = True;
                	GetColors()->add(color);
			n++;
			if (n > 543) {
				error("%s, line %d: "
				      "too many user defined colors"
					__FILE__, __LINE__);
				break;
			}
		}
		
	}
}

void FigGrafport::SetBackgroundColor(const char *c) {
        string s = c;
        SetBackgroundColor(&s);
}
 
void FigGrafport::SetForegroundColor(const char *c) {
        string s = c;
        SetForegroundColor(&s);
}
 
void FigGrafport::SetBackgroundColor(const string *c) {
	if (*GetBackgroundColor() == *c)
		return;
	Grafport::SetBackgroundColor(c);
}

void FigGrafport::SetForegroundColor(const string *c) {
	if (*GetForegroundColor() == *c)
		return;
	Grafport::SetForegroundColor(c);
	Color *color = LookupColor(c);
	if (color)
		penColor = (int)color->pixel;
	else
		penColor = -1; // default.
}

void FigGrafport::Header(bool landscape) {
	fprintf(fd, "#FIG 3.1\n");
	if (landscape)
		fprintf(fd, "Landscape\n");
	else
		fprintf(fd, "Portrait\n");
	fprintf(fd, "Center\n");
	fprintf(fd, "Inches\n");
	fprintf(fd, "1200 2\n");

	// write user defined colors.
	int object_code = 0; // color pseudo object code.
	List<Color *> *c = GetColors();
	for (c->first(); !c->done(); c->next()) {
		Color *color = c->cur();
		if (color->pixel >= 32) {
			int r= int(255*color->red);
			int g= int(255*color->green);
			int b= int(255*color->blue);
			fprintf(fd, "%d %d #", object_code, (int)color->pixel);
			fprintf(fd, "%.2x%.2x%.2x\n", r, g, b);
		}
	}
}

void FigGrafport::SetLineWidth(unsigned int w) {
	if (w == GetLineWidth())
		return;
	Grafport::SetLineWidth(w);
}

void FigGrafport::SetFont(XFont *f) {
	Grafport::SetFont(f);
	SetPSFont();
	SetTexFont();
}

void FigGrafport::DrawFigEllipse(int center_x, int center_y, int radius_x,
				 int radius_y, int sub_type) {
	int object_code = 1;  // Ellipse

	int line_style = 0;  // Solid
	if (GetLineStyle()==LineStyle::DASHED)
	 	line_style = 1;
	if (GetLineStyle()==LineStyle::DOTTED ||
	    GetLineStyle()==LineStyle::WIDE_DOTTED)
	 	line_style = 2;

	int thickness = GetLineWidth();  // 1/80 inch
	if (GetLineStyle()==LineStyle::INVISIBLE &&
	    GetFillStyle()==FillStyle::UNFILLED)
		return; // don't draw
	
	int pen_color = penColor;
	int fill_color = 7; // 7 = white
	int depth = 0;
	int pen_style = 0;  // not used.
	int area_fill = -1; // no fill.
	if (GetFillStyle()==FillStyle::FILLED) {
		fill_color = fillColor;
		area_fill = 20;// full saturation.
		if (GetLineStyle()==LineStyle::INVISIBLE)
			pen_color = fill_color;
	} 

	float style_val = 0;// 1/80 inch
	if (GetLineStyle()==LineStyle::DASHED)
	 	style_val = 4.0;
	if (GetLineStyle()==LineStyle::DOTTED)
	 	style_val = 1.0;
	if (GetLineStyle()==LineStyle::WIDE_DOTTED)
	 	style_val = 2.0;

	int direction = 1; // always 1
	float angle = 0;   // angle with x-axis.
	int start_x = center_x;
	int start_y = center_y;
	int end_x = start_x + radius_x;
	int end_y = start_y + radius_y;

	fprintf(fd, "%d %d %d %d %d %d %d %d %d ", object_code, sub_type,
		line_style, thickness, pen_color, fill_color, depth,
		pen_style, area_fill);
	fprintf(fd, "%f %d %f %d %d %d %d %d %d %d %d\n", style_val,
		direction, angle, center_x, center_y, radius_x,
		radius_y, start_x, start_y, end_x, end_y);
}

void FigGrafport::DrawFigPolyLine(List<Point *> *points, 
				  int sub_type, int radius) {
	int object_code = 2; // Polyline

	int line_style = 0;  // Solid
	if (GetLineStyle()==LineStyle::DASHED)
	 	line_style = 1;
	if (GetLineStyle()==LineStyle::DOTTED ||
	    GetLineStyle()==LineStyle::WIDE_DOTTED)
	 	line_style = 2;

	int thickness = GetLineWidth();  // 1/80 inch
	if (GetLineStyle()==LineStyle::INVISIBLE &&
            GetFillStyle()==FillStyle::UNFILLED)
		return;
	 	// thickness = 0;

	int pen_color = penColor;
	int fill_color = 7; // 7 = white
	int depth = 0;
	int pen_style = 0;  // not used.
	int area_fill = -1; // no fill.
	if (GetFillStyle()==FillStyle::FILLED) {
		fill_color = fillColor;
		area_fill = 20;// full saturation.
		if (GetLineStyle()==LineStyle::INVISIBLE)
			pen_color = fill_color;
	}

	float style_val = 0;// 1/80 inch
	if (GetLineStyle()==LineStyle::DASHED)
	 	style_val = 4.0;
	if (GetLineStyle()==LineStyle::DOTTED)
	 	style_val = 1.0;
	if (GetLineStyle()==LineStyle::WIDE_DOTTED)
	 	style_val = 2.0;

	int join_style = 0; // Miter
	int cap_style = 0;  // Butt
	int forward_arrow = 0;
	int backward_arrow = 0;
	int npoints = points->count();

	fprintf(fd, "%d %d %d %d %d %d %d %d %d %.1f %d %d %d %d %d %d\n",
		object_code, sub_type, line_style, thickness, pen_color,
		fill_color, depth, pen_style, area_fill, style_val,
		join_style, cap_style, radius, forward_arrow, backward_arrow,
		npoints);
	for (points->first(); !points->done(); points->next()) {
		Point *pt = points->cur();
		fprintf(fd, "%d %d ", pt->x, pt->y);
	}
	fprintf(fd, "\n"); 
}

void FigGrafport::DrawFigRectangle(int _x, int _y, int _wd, int _ht,
				   int sub_type, int radius) {
	int x = Zoom(_x); 
	int y = Zoom(_y);
	int wd = Zoom(_wd);
	int ht = Zoom(_ht);

	List<Point *> points;

	int x1 = int(x*F);
	int y1 = int(y*F);
	int x2 = int((x+wd)*F);
	int y2 = int(y*F);
	int x3 = int((x+wd)*F);
	int y3 = int((y+ht)*F);
	int x4 = int(x*F);
	int y4 = int((y+ht)*F);
	Point p1 = Point(x1, y1);
	Point p2 = Point(x2, y2);
	Point p3 = Point(x3, y3);
	Point p4 = Point(x4, y4);
	Point p5 = Point(x1, y1);
	points.add(&p1);
	points.add(&p2);
	points.add(&p3);
	points.add(&p4);
	points.add(&p5);

	DrawFigPolyLine(&points, sub_type, radius);

	if (GetLineStyle()==LineStyle::DUAL) {
		SetLineStyle(LineStyle::SOLID); // avoid recursion
	 	int n = 2*GetLineWidth();
		if (_ht > 2*n && _wd > 2*n)
	 		DrawFigRectangle(_x+n, _y+n, _wd-2*n, _ht-2*n, 
				sub_type, radius);
		SetLineStyle(LineStyle::DUAL);
	}
}

void FigGrafport::DrawRectangle(int x, int y, int wd, int ht) {
	DrawFigRectangle(x, y, wd, ht, 2);
	SetFillStyle(FillStyle::UNFILLED);
}

void FigGrafport::DrawUnzoomedRectangle(int x, int y, int wd, int ht) {
	DrawFigRectangle(ZoomCorrect(x), ZoomCorrect(y),
		      ZoomCorrect(wd), ZoomCorrect(ht), 2);
}

void FigGrafport::FillRectangle(int, int, int, int) {
	SetFillStyle(FillStyle::FILLED);
	SetFillColor();
}

void FigGrafport::DrawRoundedRectangle(int x, int y, int wd, int ht, int rd) {
	DrawFigRectangle(x, y, wd, ht, 4, rd);
	SetFillStyle(FillStyle::UNFILLED);
}

void FigGrafport::FillRoundedRectangle(int, int, int, int, int) {
	SetFillStyle(FillStyle::FILLED);
	SetFillColor();
}

void FigGrafport::DrawEllipsedRectangle(int x, int y, int wd, int ht, int rd) {
	DrawRoundedRectangle(x, y, wd, ht, rd);
}

void FigGrafport::FillEllipsedRectangle(int, int, int, int, int) {
	SetFillStyle(FillStyle::FILLED);
	SetFillColor();
}

void FigGrafport::DrawDisk(int x, int y, int wd, int ht, int el) {
	// draw two lines (without the corner points).
        int ah = el/2;
        DrawLine(x, y+ah+1, x, y+ht-ah-1);
        DrawLine(x+wd, y+ah+1, x+wd, y+ht-ah-1);
        // draw 1st ellipse.
        DrawEllipse(x, y, wd, el+1);
        // draw 2nd ellipse
        DrawEllipse(x, y+ht-el, wd, el+1);
}

void FigGrafport::FillDisk(int, int, int, int, int) {
	SetFillStyle(FillStyle::FILLED);
	SetFillColor();
}

void FigGrafport::DrawFigText(int px, int py, const char *str, int sub_type) {
	px = Zoom(px);
	py = Zoom(py);
	int object_code = 4;  // Text
	int color = penColor;
	int depth = 0;
	int pen_style = 0;  // not used.
	int figfont;
	if (texFonts)
		figfont = texFont; // generate LaTeX fonts.
	else
		figfont = psFont;  // generate PostScript fonts.
	float font_size = GetFont()->GetSize() - 0.1; // possibly some 
	float angle = 0; 	    // problems with 'autoresize'.
	int font_flags;
	if (texFonts)
		font_flags = 2;
	else
		font_flags = 4;
	double theight, tlength;
	GetFigStringSize(str, theight, tlength);
	int x, y;
	if (sub_type == 0) {	  // lower left (was topleft)
		x = int(F*px);
		y = int(F*py);
	}
	else if (sub_type == 1) { // lower center (was center)
		x = int(F*px);
		y = int(F*py + theight/2);
	}
	else {   		 // lower right (was topright)
		x = int(F*px);
		y = int(F*py);
	}
	string theString = str;
	MakeFigString(&theString);
	fprintf(fd, "%d %d %d %d %d %d %f %f %d %f %f %d %d %s\n",
		object_code, sub_type, color, depth, pen_style,
		figfont, font_size, angle, font_flags, theight,
		tlength, x, y, theString.getstr());
}

void FigGrafport::DrawStringLeft(int x, int y, const char *str) {
	DrawFigText(x, y, str, 0);
}

void FigGrafport::DrawUnzoomedStringLeft(int x, int y, const char *str) {
	DrawStringLeft(ZoomCorrect(x), ZoomCorrect(y), str);
}

void FigGrafport::DrawStringRight(int x, int y, const char *str) {
	DrawFigText(x, y, str, 2);
}

void FigGrafport::DrawStringCentered(int x, int y, const char *str) {
	DrawFigText(x, y, str, 1);
}

void FigGrafport::DrawUnzoomedStringCentered(int x, int y, const char *str) {
	DrawStringCentered(ZoomCorrect(x), ZoomCorrect(y), str);
}

// Underlined Strings
void FigGrafport::DrawStringRightUnderlined(int x, int y, const char *str) {
	DrawStringRight(x, y, str);
	int wd = GetFont()->StringWidth(str);
	// DrawStringRight... : 
	// Corrected Height for y ( == topRighty + GetAscent()) 
	int cht = GetFont()->GetHeight() - GetFont()->GetAscent();
	LineStyle::Type save = GetLineStyle();	// save
	SetLineStyle(LineStyle::SOLID);		// set underline linestyle SOLID
	DrawLine(x-wd, y+cht, x, y+cht);
	SetLineStyle(save);			// and restore
}

void FigGrafport::DrawStringLeftUnderlined(int x, int y, const char *str) {
	DrawStringLeft(x, y, str);
	int wd = GetFont()->StringWidth(str);
	// DrawStringLeft... : 
	// Corrected Height for y ( == topLefty + GetAscent()) 
	int cht = GetFont()->GetHeight() - GetFont()->GetAscent();
	LineStyle::Type save = GetLineStyle();	// save
	SetLineStyle(LineStyle::SOLID);		// set underline linestyle SOLID
	DrawLine(x, y+cht, x+wd, y+cht);
	SetLineStyle(save);			// and restore
}

void FigGrafport::DrawStringCenteredUnderlined(int x, int y, const char *str) {
	DrawStringCentered(x, y, str);
	int wd = GetFont()->StringWidth(str);
	int ht = GetFont()->GetHeight();
	LineStyle::Type save = GetLineStyle();	// save
	SetLineStyle(LineStyle::SOLID);		// set underline linestyle SOLID
	DrawLine(x-wd/2, y+ht/2, x+wd/2, y+ht/2);	
	SetLineStyle(save);			// and restore
}

void FigGrafport::DrawLine(int _x1, int _y1, int _x2, int _y2) {
	if (_x1 == _x2 && _y1 == _y2)
		return;

	int x1 = Zoom(_x1); 
	int y1 = Zoom(_y1);
	int x2 = Zoom(_x2);
	int y2 = Zoom(_y2);

	List<Point *> points;
	int sub_type = 1;

	Point p1 = Point(int(F*x1), int(F*y1));
	Point p2 = Point(int(F*x2), int(F*y2));
	points.add(&p1);
	points.add(&p2);

	DrawFigPolyLine(&points, sub_type);

	if (GetLineStyle()==LineStyle::DUAL) {
		SetLineStyle(LineStyle::SOLID);
		int n = GetLineWidth();
		if (y1 == y2) {  // horizontal
			DrawLine(_x1, _y1+n, _x2, _y2+n);
			DrawLine(_x1, _y1-n, _x2, _y2-n);
		}
		else if (x1 == x2) {  // vertical
			DrawLine(_x1+n, _y1, _x2+n, _y2);
			DrawLine(_x1-n, _y1, _x2-n, _y2);
		}
		else {
			DrawLine(_x1, _y1, _x2, _y2);
			DrawLine(_x1, _y1+2*n, _x2, _y2+2*n);
		}
		SetLineStyle(LineStyle::DUAL);
	}
	SetFillStyle(FillStyle::UNFILLED);
}

void FigGrafport::DrawUnzoomedLine(int x1, int y1, int x2, int y2) {
	DrawLine(ZoomCorrect(x1), ZoomCorrect(y1), 
		 ZoomCorrect(x2), ZoomCorrect(y2));
}

void FigGrafport::DrawCurve(const Point *p1, const Point *p2, 
			   const Point *p3, const Point *p4) {
	int object_code = 3;  // Spline
	int sub_type = 0;     // Open Aproximated Spline
	
	int x1 = Zoom(int(p1->x*F));
	int y1 = Zoom(int(p1->y*F));
	int x2 = Zoom(int(p2->x*F));
	int y2 = Zoom(int(p2->y*F));
	int x3 = Zoom(int(p3->x*F));
	int y3 = Zoom(int(p3->y*F));
	int x4 = Zoom(int(p4->x*F));
	int y4 = Zoom(int(p4->y*F));

	int line_style = 0;  // Solid
	if (GetLineStyle()==LineStyle::DASHED)
	 	line_style = 1;
	if (GetLineStyle()==LineStyle::DOTTED ||
	    GetLineStyle()==LineStyle::WIDE_DOTTED)
	 	line_style = 2;

	int thickness = GetLineWidth();  // 1/80 inch
	if (GetLineStyle()==LineStyle::INVISIBLE)
		return;
	 	// thickness = 0;

	int pen_color = penColor;
	int fill_color = 7; // 7 = white
	int depth = 0;
	int pen_style = 0;  // not used.
	int area_fill = -1; // no fill.
	if (GetFillStyle()==FillStyle::FILLED) {
		fill_color = fillColor;
		area_fill = 20;// full saturation.
	} 

	float style_val = 0;// 1/80 inch
	if (GetLineStyle()==LineStyle::DASHED)
	 	style_val = 4.0;
	if (GetLineStyle()==LineStyle::DOTTED)
	 	style_val = 1.0;
	if (GetLineStyle()==LineStyle::WIDE_DOTTED)
	 	style_val = 2.0;

	int cap_style = 0;  // Butt
	int forward_arrow = 0;
	int backward_arrow = 0;
	int npoints = 4;

	fprintf(fd, "%d %d %d %d %d %d %d %d %d %.1f %d %d %d %d\n",
		object_code, sub_type, line_style, thickness, pen_color,
		fill_color, depth, pen_style, area_fill, style_val,
		cap_style, forward_arrow, backward_arrow,
		npoints);
	fprintf(fd, "%d %d %d %d %d %d %d %d\n", 
		x1, y1, x2, y2, x3, y3, x4, y4);
	SetFillStyle(FillStyle::UNFILLED);
}

void FigGrafport::DrawPoint(int, int) { 
	// no single points in Xfig
}

void FigGrafport::DrawArc(int x, int y, int wd, int ht, int a1, int a2) {
	double deg2rad = 0.017453293;
	a1 = a1%360;
	a2 = a2%360;
	// cout << "a1 = " << a1 << " a2 = " << a2 << endl;
	double arc1 = a1 * deg2rad;
	double arc2 = a2 * deg2rad;
	double arc3, arc4;
	if (arc1 < arc2) {
		arc3 =  deg2rad * (2*a1 + a2) / 3;
		arc4 =  deg2rad * (a1 + 2*a2) / 3;
	}
	else {
		arc3 = deg2rad * (2*a1 + (360+a2)) / 3;
		arc4 = deg2rad * (a1 + 2*(360+a2)) / 3;
	}
	// cout << "arc1 = " << arc1 << " arc2 = " << arc2 << endl;
	// cout << "arc3 = " << arc3 << " arc4 = " << arc4 << endl;
	Point p0 = Point(x, y);
	Point t1, t2, t3, t4;
	t1.x = p0.x + (int) (100000 * sin(arc1));
	t1.y = p0.y + (int) (100000 * cos(arc1));
	t2.x = p0.x + (int) (100000 * sin(arc2));
	t2.y = p0.y + (int) (100000 * cos(arc2));
	t3.x = p0.x + (int) (100000 * sin(arc3));
	t3.y = p0.y + (int) (100000 * cos(arc3));
	t4.x = p0.x + (int) (100000 * sin(arc4));
	t4.y = p0.y + (int) (100000 * cos(arc4));
	// cout << "p0 = " << p0 << endl;
	// cout << "width = " << wd << " height = " << ht << endl;
	// cout << "t1 = " << t1 << " t2 = " << t2 << endl;
	// cout << "t3 = " << t3 << " t4 = " << t4 << endl;
	Point p1 = Intersect::EllipseIntersects(p0, t1, wd, ht);
	Point p2 = Intersect::EllipseIntersects(p0, t2, wd, ht);
	Point p3 = Intersect::EllipseIntersects(p0, t3, wd, ht);
	Point p4 = Intersect::EllipseIntersects(p0, t4, wd, ht);
	// cout << "p1 = " << p1 << " p2 = " << p2 << endl;
	// cout << "p3 = " << p3 << " p4 = " << p4 << endl;
	DrawCurve(&p1, &p3, &p4, &p2);
}

void FigGrafport::FillArc(int, int, int, int, int, int) {
	SetFillStyle(FillStyle::FILLED);
	SetFillColor();
}

void FigGrafport::DrawEllipse(int x, int y, int wd, int ht) {
	int sub_type;
	if (wd == ht)
		sub_type = 3; // circle def. by radius.
	else
		sub_type = 1; // ellipse def. by radius.
	int radius_x = int((wd*F)/2);
	int radius_y = int((ht*F)/2);
	int center_x = int(x*F + radius_x);
	int center_y = int(y*F + radius_y);
	DrawFigEllipse(center_x, center_y, radius_x, radius_y, sub_type);
	if (GetLineStyle()==LineStyle::DUAL) {
		SetLineStyle(LineStyle::SOLID); // avoid recursion
	 	int n = 2*GetLineWidth();
		if (ht > 2*n && wd > 2*n)
	 		DrawEllipse(x+n, y+n, wd-2*n, ht-2*n);
		SetLineStyle(LineStyle::DUAL);
	}
	SetFillStyle(FillStyle::UNFILLED);
}

void FigGrafport::FillEllipse(int, int, int, int) {
	SetFillStyle(FillStyle::FILLED);
	SetFillColor();
}

void FigGrafport::DrawPolygon(const Point *p, int n) {
	if (n < 1) 
		return;
	List<Point *> points;
	int sub_type = 3;

	for (int i=0; i<n; i++) {
		int x = int(F*Zoom(p[i].x)); 
		int y = int(F*Zoom(p[i].y));
		Point *pt = new Point(x, y);
		points.add(pt);
	}
	Point *pt = new Point(
			int(F*Zoom(p[0].x)), int(F*Zoom(p[0].y)));
	points.add(pt);

	DrawFigPolyLine(&points, sub_type);
	if (GetLineStyle()==LineStyle::DUAL) {
		int w = 2*GetLineWidth();
                Point *oldPoints = new Point[n];
                Point *newPoints = new Point[n];
		for (int j=0; j<n; j++)
			oldPoints[j] = *points[j];
                RecalcPolygon(oldPoints, n, newPoints, w);
		List<Point *> nPoints;
		for (int k=0; k<n; k++)
			nPoints.add(new Point(newPoints[k]));
                DrawFigPolyLine(&nPoints, sub_type);
                delete newPoints;
                delete oldPoints;
		nPoints.clear();
	}
	points.clear();
	SetFillStyle(FillStyle::UNFILLED);
}

void FigGrafport::FillPolygon(const Point *, int) {
	SetFillStyle(FillStyle::FILLED);
	SetFillColor();
}

void FigGrafport::MakeFigString(string *str) {
	char buf[8];
	string newstr;
	for (unsigned i=0; i<str->length(); i++) {
		unsigned char c = (*str)[i];
		if (!isascii(c)) {
			sprintf(buf, "\\%o", c);
			newstr += buf;
		}
		else if (c == '\\') {
			newstr.add(c);
			newstr.add(c);
		}
		else
			newstr.add(c);
	}
	newstr += "\\001";
	*str = newstr;
}

void FigGrafport::SetFillColor() {
	Color *color = LookupColor(GetForegroundColor());
        if (color)
                fillColor = (int)color->pixel;
        else
                fillColor = -1; // default.
}

void FigGrafport::GetFigStringSize(const char *str, double &ht, double &wd) {
	XCharStruct overall;
	int dir, asc, desc;
	int nchars = strlen(str);
	XTextExtents(GetFont()->GetXFontStruct(),
		(char *)str, nchars, &dir, &asc, &desc, &overall);
	ht = F * (overall.ascent + overall.descent);
	wd = F * overall.width;
}

void FigGrafport::SetPSFont() {
	XFont *f = GetFont();
	int fam = f->GetFamily();
	if (fam == XFont::HELVETICA)
		psFont = 16;
	else if (fam == XFont::TIMESROMAN)
		psFont = 0;
	else if (fam == XFont::COURIER)
		psFont = 12;
	else if (fam == XFont::NEWCENTURYSCHLBK)
		psFont = 24;
	else if (fam == XFont::SYMBOL) {
		psFont = 32;
		return;
	}
	if (f->IsItalic())
		psFont += 1;
	if (f->IsBold())
		psFont += 2;
}

void FigGrafport::SetTexFont() {
	XFont *f = GetFont();
	texFont = 1;
	if (f->IsBold())
		texFont += 1;
	if (f->IsItalic())
		texFont += 2;
	int fam = f->GetFamily();
	if (fam == XFont::COURIER)
		texFont = 5;
}
