////////////////////////////////////////////////////////////////////////////////
//
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1995, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////
#include "grafport.h"
#include "application.h"
#include "llist.h"
#include <math.h>

Grafport::Grafport() {
	zoomValue = 1.0;
	int f = XFont::HELVETICA;
	int s = XFont::PLAIN;
	int p = 12;
	string foundry = "Adobe";
	defaultFont = new XFont(&foundry, f, s, p, False);
	font = defaultFont;
	backgroundColor = "white";
	foregroundColor = "black";
	lineWidth = 1;
	lineStyle = LineStyle::SOLID;
	fillStyle = FillStyle::UNFILLED;
	colors = new List<Color *>;
}

Grafport::~Grafport () { 
	delete defaultFont;
	colors->clear();
	delete colors;
}

void Grafport::InitColors() {
	List<Color *> *appColors = theApplication->GetColors();
        for (appColors->first(); !appColors->done(); appColors->next()) {
                Color *color = new Color(*appColors->cur());
		color->pixel = 0;
		color->allocated = False;
		colors->add(color);
	}
}

unsigned Grafport::CursorPosition(int x, int y, const string *str, 
		int hit_x, int hit_y) {
	int j = 0; // index in string.
	unsigned n = 1; // number of lines processed.
	unsigned nLines = str->lines();
	// the cursor position, closest to hit_x, hit_y.
	unsigned cursor = 0;
	int fontHeight = font->GetHeight();
	int offset = y - (nLines) * fontHeight/2;
	buf = "";
	const char *s = str->getstr();
	// string is non-zero.
	while (n <= nLines) {
		char c = s[j];
		buf.add(c);
		// end of line ? 
		if (c == '\n' || c == '\r' || c == '\0') {
			buf.remove();
			int y1 = offset + (n-1) * fontHeight;
			int y2 = offset + n * fontHeight;
			if (y1 <= hit_y && hit_y <= y2) {
				// found right line, look at hit_x;
				Point p1 = CursorPosition(x,y,&buf,0);
				for (unsigned k = 1; k <= buf.length(); k++) {
					Point p2 = CursorPosition(x,y,&buf,k);
					int mid = (p1.x + p2.x)/2;
					if (hit_x <= mid)
						return min(cursor, 
							   str->length());
					else {
						p1 = p2;
						cursor++;
					}
				}
				return min(cursor, str->length());
			}
			else
				cursor += buf.length()+1;
			buf = ""; // another line
			n++;
		}
		j++;
	}
	return min(cursor, str->length());
}

Point Grafport::DrawStringsCentered(int x, int y, const string *str, 
		bool ul, int cursor) {
	int j = 0; // index in string.
	unsigned n = 1; // number of lines processed.
	unsigned nLines = str->lines();
	Point p = Point(x,y);   // cursor position.
	int fontHeight = font->GetHeight();
	int offset = y - (nLines-1) * fontHeight/2;
	buf = "";
	const char *s = str->getstr();
	// string is non-zero.
	while (n <= nLines) {
		char c = s[j];
		buf.add(c);
		// end of line ? 
		if (c == '\n' || c == '\r' || c == '\0') {
			buf.remove();
			// draw string in buf
			y = offset + (n-1) * fontHeight;
			if (ul) 
				DrawStringCenteredUnderlined(x, y, buf.getstr());
			else
				DrawStringCentered(x, y, buf.getstr());
			// look for cursor position.
			if (cursor <= j && cursor >= (int)(j - buf.length()))
				p = CursorPosition(x, y, &buf,
						cursor-j+buf.length());
			buf = ""; // another line
			n++;
		}
		j++;
	}
	return p;
}

Point Grafport::DrawStringsLeft(int topLeftx, int topLefty, 
		const string *str, bool ul, int cursor) {
	int j = 0; // index in string.
	unsigned n = 1; // number of lines processed.
	unsigned nLines = str->lines();
	topLefty += font->GetAscent();
	Point p = Point(topLeftx,topLefty);   // cursor position.
	int fontHeight = font->GetHeight();
	buf = "";
	const char *s = str->getstr();
	// string is non-zero.
	while (n <= nLines) {
		char c = s[j];
		buf.add(c);
		// end of line ? 
		if (c == '\n' || c == '\r' || c == '\0') {
			buf.remove();
			// draw string in buf
			if (ul) 
				DrawStringLeftUnderlined(
					topLeftx, topLefty, buf.getstr());
			else
				DrawStringLeft(
					topLeftx, topLefty, buf.getstr());
			// look for cursor position.
			if (cursor <= j && cursor >= (int)(j - buf.length()))
				p = CursorPosition(
					topLeftx+font->StringWidth(&buf)/2,
					topLefty-font->GetDescent(), &buf,
					cursor-j+buf.length());
			topLefty += fontHeight;
			buf = ""; 	// another line
			n++;
		}
		j++;
	}
	return p;
}

Point Grafport::DrawStringsRight(int topRightx, int topRighty, 
		const string *str, bool ul, int cursor) { 
	int j = 0; // index in string.
	unsigned n = 1; // number of lines processed.
	unsigned nLines = str->lines();
	topRighty += font->GetAscent();
	Point p = Point(topRightx,topRighty);   // cursor position.
	int fontHeight = font->GetHeight();
	buf = "";
	const char *s = str->getstr();
	// string is non-zero.
	while (n <= nLines) {
		char c = s[j];
		buf.add(c);
		// end of line ? 
		if (c == '\n' || c == '\r' || c == '\0') {
			buf.remove();
			// int x = topRightx - font->StringWidth(&buf);
			// draw string in buf
			if (ul) 
				DrawStringRightUnderlined(
					topRightx, topRighty, buf.getstr());
			else
				DrawStringRight(
					topRightx, topRighty, buf.getstr());
			// look for cursor position.
			if (cursor <= j && cursor >= (int)(j - buf.length()))
				p = CursorPosition(
					topRightx-font->StringWidth(&buf)/2, 
					topRighty-font->GetDescent(), &buf,
					cursor-j+buf.length());
			topRighty += fontHeight;
			buf = ""; 	// another line
			n++;
		}
		j++;
	}
	return p;
}

Point Grafport::CursorPosition(int x, int y, const string *text, int cursor) {
	string str;
	for(int i=0; i<cursor; i++)
		str.add((*text)[i]);
	int width = font->StringWidth(&str);
	int base = font->GetHeight()/2 - font->GetDescent();
	int str_x;
	if (font->IsScalable())
		str_x = x - font->StringWidth(text)/2 + width;
	else
		str_x = x - ZoomCorrect(font->StringWidth(text)/2 - width);
	return Point(str_x, y+base);
}

void Grafport::DrawRectangles(const Rectangle *rects, int n){
	for (int i=0; i<n; i++)
		DrawRectangle(rects[i].x,rects[i].y,
				rects[i].width,rects[i].height);
}

void Grafport::FillRectangles(const Rectangle *rects, int n){
	for (int i=0; i<n; i++)
		FillRectangle(rects[i].x,rects[i].y,
				rects[i].width,rects[i].height);
}

Color *Grafport::LookupColor(const string *colorName) {
        for (colors->first(); !colors->done(); colors->next()) {
                Color *color = colors->cur();
                if (color->name %= *colorName)
                        return color;
        }
        return 0;
}

Color *Grafport::LookupColor(const char *colorName) {
	string x = colorName;
	return LookupColor(&x);
}

void Grafport::AddColor(Color *c) {
	colors->add(new Color(*c));
}

void Grafport::AddColor(const string *name, unsigned long pixel) {
	Color *color = new Color(*name, pixel);
        colors->add(color);
}

void Grafport::AddColor(const char *name, unsigned long pixel) {
	string x = name;
	AddColor(&x, pixel);
}

void Grafport::RecalcPolygon(const Point *points, int nrPoints, Point *newPoints, int d) {
	// it is a regular polygon.
	Point p0 = points[0];
	Point pN;
	if (nrPoints % 2 == 0)
		pN = points[nrPoints/2];
	else {
		Point pN1 = points[nrPoints/2];
		Point pN2 = points[1+nrPoints/2];
		pN.x = (pN1.x + pN2.x) / 2;
		pN.y = (pN1.y + pN2.y) / 2;
	}
	Point pM;
	pM.x = (p0.x + pN.x) / 2;
	pM.y = (p0.y + pN.y) / 2;
	for (int i = 0; i < nrPoints; i++) {
		Point p = points[i];
		if (p.x < pM.x)
			p.x += d;
		else if (p.x > pM.x)
			p.x -= d;
		if (p.y < pM.y)
			p.y += d;
		else if (p.y > pM.y)
			p.y -= d;
		newPoints[i] = p;
	}
}
