//------------------------------------------------------------------------------
// 
// This file is part of Toolkit for Conceptual Modeling (TCM).
// (c) copyright 1995, Vrije Universiteit Amsterdam.
// Author: Frank Dehne (frank@cs.vu.nl).
//
// TCM is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// TCM is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with TCM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
//-----------------------------------------------------------------------------
#ifndef _GRAFPORT_H 
#define _GRAFPORT_H

#include "rectangle.h"
#include "lstring.h"
#include "xfont.h"
#include "linestyle.h"
#include "fillstyle.h"
#include "llist.h"
#include "color.h"

/// (abstract) grafport class (primitive drawing functions).
class Grafport {
/*@Doc: {\large {\bf scope:} GUI} */
public:
	///
	Grafport();
	///
	virtual ~Grafport();

	///
	virtual bool DrawsXor() {return False;}

// Sizes, refreshment.
	///
	virtual void ClearArea(int x, int y, int wd, int ht)=0;
	///
	virtual void UpdateSize(int, int) {;}
	///
	virtual int GetWidth() {return 0;}
	///
	virtual int GetHeight() {return 0;}
	///
	virtual void Redraw(int, int, int, int, int, int) {;}

// Fonts
	/// Make this the current font.
	virtual void SetFont(XFont *f) {font = f;}
	/// Get the current font.
	XFont *GetFont() {return font;}

// Colors
	///
	virtual void SetBackgroundColor(const char *colorName)
		{backgroundColor = colorName;}
	///
	virtual void SetBackgroundColor(const string *colorName)
		{backgroundColor = *colorName;}
	///
	virtual void SetForegroundColor(const char *colorName)
		{foregroundColor = colorName;}
	///
	virtual void SetForegroundColor(const string *colorName) 
		{foregroundColor = *colorName;}
	///
	const string *GetBackgroundColor() {return &backgroundColor;}
	///
	const string *GetForegroundColor() {return &foregroundColor;}

	///
	List<Color *> *GetColors() {return colors;}

// Line styles.
	///
	void SetLineStyle(LineStyle::Type s) {lineStyle=s;}
	///
	LineStyle::Type GetLineStyle() {return lineStyle;}

// Fill styles.
	///
	void SetFillStyle(FillStyle::Type s) {fillStyle=s;}
	///
	FillStyle::Type GetFillStyle() {return fillStyle;}

// Line width.
	///
	virtual void SetLineWidth(unsigned n) {lineWidth=n;}
	///
	unsigned GetLineWidth() {return lineWidth;}

/// Zooming in/out.
	/// set zoom value.
	void SetZoomValue(double newValue) {zoomValue = newValue;}
	///
	double GetZoomValue() {return zoomValue;}
	///
	int Zoom(int x) {return (int) (x * zoomValue);}
	///
	int ZoomCorrect(int x) {return (int) (x / zoomValue);}

	// All draw functions, except those called 'Edit', are printable
	// (are exported to Postscript etc.)
	// All draw functions, except those called 'Edit' or 'Unzoomed' are 
	// zoomed.

// Drawing Rectangles.
	///
	virtual void DrawRectangle(int x, int y, int wd, int ht)=0;
	///
	virtual void FillRectangle(int x, int y, int wd, int ht)=0;

	///
	virtual void DrawRoundedRectangle(
		int x, int y, int wd, int ht, int rd)=0;
	///
	virtual void FillRoundedRectangle(
		int x, int y, int wd, int ht, int rd)=0;

	///
        virtual void DrawEllipsedRectangle(
                int x, int y, int wd, int ht, int rd)=0;
        ///
        virtual void FillEllipsedRectangle(
                int x, int y, int wd, int ht, int rd)=0;
	
	/// Draw list of n rectangles.
	void DrawRectangles(const Rectangle *rects, int n);
	///
	void FillRectangles(const Rectangle *rects, int n);

	/// Draw dashed rectangle (not zoomed, not printed)
	virtual void DrawEditDashedRectangle(int, int, int, int) {;}
	///
	virtual void DrawUnzoomedRectangle(int, int, int, int) = 0;

// Drawing Strings
	/// Draw string with topleft at (x,y)
	virtual void DrawStringLeft(int x, int y, const char *str)=0;

	/// Draw string centered at (x,y)
	virtual void DrawStringCentered(int x, int y, const char *str)=0;

	/// Draw string with topright at (x,y)
	virtual void DrawStringRight(int x, int y, const char *str)=0;

	/// Draw string with topleft underlined at (x,y)
	virtual void DrawStringLeftUnderlined(int x, int y, const char *str)=0;

	/// Draw string centered underlined at (x,y)
	virtual void DrawStringCenteredUnderlined(
		int x, int y, const char *str)=0;

	/// Draw string with topright underlined at (x,y)
	virtual void DrawStringRightUnderlined(int x, int y, const char *str)=0;

	/// Draws n-line string centered (underlined?) at (x,y),
	/// returns cursor position.
	Point DrawStringsCentered(int x, int y, const string *str,
					bool ul, int cursor=0);

	/// draw n-line string left aligned, returns cursor position.
	Point DrawStringsLeft(
			int topLeftx, int topLefty, 
			const string *str, bool ul, int cursor=0);

	/// draw n-line string right aligned, return cursor position.
	Point DrawStringsRight(
			int topRightx, int topRighty, 
			const string *str, bool ul, int cursor=0);

	/// return cursor place closest to hitX, hitY.
	unsigned CursorPosition(
			int x, int y, const string *str, 
			int hitX, int hitY);

	///
	virtual void DrawUnzoomedStringLeft(int, int, const char *)=0;
	///
	virtual void DrawUnzoomedStringCentered(int, int, const char *)=0;

// Drawing Lines
	///
	virtual void DrawLine(const Point *p1, const Point *p2)=0;
	///
	virtual void DrawLine(int x1, int y1, int x2, int y2)=0;

	/// 
	virtual void DrawCurve(const Point *p1, const Point *p2, 
			const Point *p3, const Point *p4)=0;

	/// not zoomed not printed.
	virtual void DrawEditLine(int, int, int, int) {;}

	/// not zoomed not printed.
	virtual void DrawEditLine(const Point *, const Point *) {;}

	/// not zoomed or printed. Uses special dot pattern.
	virtual void DrawEditDottedGridLine(int, int, int, int) {;}
	///
	virtual void DrawEditDottedLine(int, int, int, int) {;}

	/// not zoomed or printed.
	virtual void DrawEditPoint(int, int) {;}

	/// not zoomed or printed.
	virtual void DrawEditPoint(const Point *) {;}

	///
	virtual void DrawUnzoomedLine(int, int, int, int) = 0;

// Drawing Points
	///
	virtual void DrawPoint(int x, int y)=0;
	///
	virtual void DrawPoint(const Point *p)=0;

// Drawing Ellipses
	///
	virtual void DrawEllipse(int x, int y, int wd, int ht)=0;
	/// 
	virtual void FillEllipse(int x, int y, int wd, int ht)=0;

// Drawing Arcs
	///
	virtual void DrawArc(int x, int y, int wd, int ht, 
				int arc1, int arc2)=0;
	/// 
	virtual void FillArc(int x, int y, int wd, int ht,
				int arc1, int arc2)=0;

// Drawing Disks
	/// disk is an ellipse, two vert. lines and a half ellipse
        virtual void DrawDisk(
                int x, int y, int wd, int ht, int ellipse_ht)=0;
        ///
        virtual void FillDisk(
                int x, int y, int wd, int ht, int ellipse_ht)=0;
	

// Drawing Polygons
	/// draw lines between the first n points in plist.
	virtual void DrawPolygon(const Point *plist, int n)=0;
	///
	virtual void FillPolygon(const Point *plist, int n)=0;

// Drawing text edit cursors.
	/// Draw active cursor sign.
	virtual void DrawActiveCaret(int, int) {;}

	/// Draw deactive cursor sign.
	virtual void DrawDeactiveCaret(int, int) {;}

protected:
	///
	Point CursorPosition(int x, int y, const string *buf, int cursor);

	///
        Color *LookupColor(const string *colorName);
	///
        Color *LookupColor(const char *colorName);
	///
	void AddColor(const string *colorName, unsigned long pixel);
	///
	void AddColor(const char *colorName, unsigned long pixel);
	///
	void AddColor(Color *c);

	///
	virtual void InitColors();

	///
	void RecalcPolygon(const Point *points, int nrPoints, 
			   Point *newPoints, int d);

private:
	/// Current font.
	XFont *font;

	/// Default font.
	XFont *defaultFont;

	/// Current background color.
	string backgroundColor;

	/// Current foreground color.
	string foregroundColor;

	/// current line width.
	unsigned lineWidth;

	/// current line style.
	LineStyle::Type lineStyle;

	/// current fill style.
	FillStyle::Type fillStyle;

	/// current zoom value.
	double zoomValue;

	///
	string buf;

	/// a cache for the colors.
        List<Color *> *colors;
};
#endif
